#cython: embedsignature=True
#cython: profile=False

import os
import sys
import numpy as np
cimport numpy as np
np.import_array()

# Set byteorder for __reduce__
byteorder = sys.byteorder


def get_include():
	"""
	Get file directory if C headers

	Parameters
	----------
		None

	Returns
	-------
		location : str
			Directory to header files
	"""

	# Grab file location
	location = os.path.split(os.path.realpath(__file__))[0]

	return location


cdef class Interval(object):
	"""
	Wrapper of C interval_t

	:class:`~Interval_core.Interval` stores an interval

	"""

	def __init__(self, start=None, end=None):
		"""
		Initialize Interval

		Parameters
		----------
			start : int
				Starting position [default = None]
			end : int
				Ending position [default = None]

		Returns
		-------
			None

		"""

		self.start = start
		self.end = end


	def __hash__(self):
		"""
		Get hash value
		"""

		return hash(repr(self))


	def __eq__(self, other):
		"""
		Check if there is overlap
		"""

		# Check that the classes match
		is_equal = self.__class__ == other.__class__

		# Check for overlap
		if is_equal:
			if other.start < self.end and other.end > self.start:
				is_equal = True
			else:
				is_equal = False

		return is_equal


	def __str__(self):
		format_string = "Interval(%d-%d)" % (self.start, self.end)
		return format_string


	def __repr__(self):
		format_string = "Interval(%d-%d)" % (self.start, self.end)
		return format_string



import numpy as np
cimport numpy as np
cimport cython
from libc.stdint cimport uint32_t, int64_t, uint8_t
from ailist.Interval_core cimport Interval, interval_t
from ailist.array_query_core cimport *

cdef extern from "src/ailist/augmented_interval_list.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_add.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_construct.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_query.c":
	# C is include here so that it doesn't need to be compiled externally
	pass
	
cdef extern from "src/ailist/ailist_iter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_get_id.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_coverage.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_nhits.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_wps.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_merge.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_extract.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_ops.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_filter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/ailist_simulate.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/overlap_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/utilities/utilities.h":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/array_query/array_query_utilities.h":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/augmented_interval_list.h":

	ctypedef struct ailist_t:
		int64_t nr, mr  					# Number of regions
		interval_t *interval_list			# Regions data
		uint32_t first, last				# Record range of intervals
		int nc
		int lenC[10]
		int idxC[10]
		uint32_t *maxE
		
	ctypedef struct ailist_sorted_iter_t:
		ailist_t *ail						# Interval list
		int nc								# Number of components
		int *comp_bounds					# Label component bounds
		int *comp_used						# Components used
		interval_t *intv					# Interval
		int n								# Current position

	ctypedef struct overlap_index_t:
		int size;							# Current size
		int max_size;						# Maximum size
		ailist_t *ail;						# Store ailist
		long *indices;						# Store indices


	#-------------------------------------------------------------------------------------
	# augmented_interval_list.c
	#=====================================================================================

	# Initialize ailist_t
	ailist_t *ailist_init() nogil

	# Free ailist data
	void ailist_destroy(ailist_t *ail) nogil

	# Print AIList
	void display_list(ailist_t *ail) nogil

	# Calculate maximum length
	int ailist_max_length(ailist_t *ail) nogil

	# Calculate length distribution
	void ailist_length_distribution(ailist_t *ail, int distribution[]) nogil


	#-------------------------------------------------------------------------------------
	# overlap_index.c
	#=====================================================================================

	# Initialize overlap_index_t
	overlap_index_t *overlap_index_init() nogil

	# Free overlap_index memory
	void overlap_index_destroy(overlap_index_t *oi) nogil

	# Add interval and index to overlap_index
	void overlap_index_add(overlap_index_t *aq, interval_t *i) nogil


	#-------------------------------------------------------------------------------------
	# ailist_add.c
	#=====================================================================================

	# Add interval to ailist_t object 
	void ailist_add(ailist_t *ail, uint32_t start, uint32_t end, uint32_t id) nogil

	# Build ailist from arrays
	void ailist_from_array(ailist_t *ail, const long starts[], const long ends[], const long ids[], int length) nogil

	# Append two ailist
	ailist_t *ailist_append(ailist_t *ail1, ailist_t *ail2) nogil

	# Copy ailist
	ailist_t *ailist_copy(ailist_t *ail) nogil


	#-------------------------------------------------------------------------------------
	# ailist_construct.c
	#=====================================================================================

	# Construct ailist: decomposition and augmentation
	void ailist_construct(ailist_t *ail, int cLen) nogil

	# Construct ailist: decomposition and augmentation v0
	void ailist_construct_v0(ailist_t *ail, int cLen) nogil

	# Validation that construction ran
	int ailist_validate_construction(ailist_t *ail) nogil

	# Calculate coverage of midpoints
	void ailist_midpoint_coverage(ailist_t *ail, double coverage[]) nogil

	# Calculate coverage of midpoints with length
	void ailist_midpoint_coverage_length(ailist_t *ail, double coverage[], int min_length, int max_length) nogil



	#-------------------------------------------------------------------------------------
	# ailist_get_id.c
	#=====================================================================================

	# Get intervals with id
	ailist_t *ailist_get_id(ailist_t *ail, int query_id) nogil

	# Get intervals with ids
	ailist_t *ailist_get_id_array(ailist_t *ail, const long ids[], int length) nogil

	# Reset id_values
	void ailist_reset_id(ailist_t *ail) nogil

	# Reset id_values with shift
	void ailist_reset_id_shift(ailist_t *ail, int shift) nogil


	#-------------------------------------------------------------------------------------
	# ailist_query.c
	#=====================================================================================

	# Binary search
	uint32_t binary_search(interval_t* As, uint32_t idxS, uint32_t idxE, uint32_t qe) nogil

	# Query ailist intervals
	void ailist_query(ailist_t *ail, ailist_t *overlaps, uint32_t qs, uint32_t qe) nogil

	# Query ailist intervals of a length
	void ailist_query_length(ailist_t *ail, ailist_t *overlaps, uint32_t qs, uint32_t qe, int min_length, int max_length) nogil

	# Query number of hits in ailist intervals
	void ailist_query_nhits(ailist_t *ail, long *nhits, uint32_t qs, uint32_t qe) nogil

	# Query number of hits in ailist intervals of a length
	void ailist_query_nhits_length(ailist_t *ail, long *nhits, uint32_t qs, uint32_t qe, int min_length, int max_length) nogil

	# Query if interval has any overlap in ailist intervals
	void ailist_query_has_hit(ailist_t *ail, uint8_t *has_hit, uint32_t qs, uint32_t qe) nogil

	# Query ailist intervals from arrays
	void ailist_query_from_array(ailist_t *ail, ailist_t *overlaps, const long starts[], const long ends[], int length) nogil

	# Query ailist intervals from another ailist
	void ailist_query_from_ailist(ailist_t *ail, ailist_t *ail2, ailist_t *overlaps) nogil

	# Query aiarray intervals and record original index
	void ailist_query_with_index(ailist_t *ail, overlap_index_t *overlaps, uint32_t qs, uint32_t qe) nogil

	# Query aiarray intervals and record original index
	void ailist_query_only_index(ailist_t *ail, array_query_t *aq, uint32_t qs, uint32_t qe, uint32_t id) nogil

	# Query ailist interval ids from array
	void ailist_query_id_from_array(ailist_t *ail, array_query_t *aq, const long starts[], const long ends[], const long ids[], int length) nogil

	# Query ailist interval ids from another ailist
	void ailist_query_id_from_ailist(ailist_t *ail, ailist_t *ail2, array_query_t *aq) nogil


	#-------------------------------------------------------------------------------------
	# ailist_iter.c
	#=====================================================================================

	# Get component index
	int *get_comp_bounds(ailist_t *ail) nogil

	# 
	ailist_sorted_iter_t *ailist_sorted_iter_init(ailist_t *ail) nogil

	# 
	int ailist_sorted_iter_next(ailist_sorted_iter_t *iter) nogil

	# 
	void ailist_sorted_iter_destroy(ailist_sorted_iter_t *iter) nogil


	#-------------------------------------------------------------------------------------
	# ailist_coverage.c
	#=====================================================================================

	# Calculate coverage for a single interval
	void ailist_interval_coverage(ailist_t *ail, int start, int end, int coverage[]) nogil

	# Calculate coverage
	void ailist_coverage(ailist_t *ail, double coverage[]) nogil

	# Calculate coverage of a length
	void ailist_coverage_length(ailist_t *ail, double coverage[], int min_length, int max_length) nogil

	# Calculate coverage within bins
	void ailist_bin_coverage(ailist_t *ail, double coverage[], int bin_size) nogil

	# Calculate coverage within bins of a length
	void ailist_bin_coverage_length(ailist_t *ail, double coverage[], int bin_size, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# ailist_nhits.c
	#=====================================================================================

	# Determine number of hits for each query interval
	void ailist_nhits_from_array(ailist_t *ail, const long starts[], const long ends[],
								int length, int nhits[]) nogil

	# Determine number of hits of a length for each query interval
	void ailist_nhits_from_array_length(ailist_t *ail, const long starts[], const long ends[],
										int length, int nhits[], int min_length,
										int max_length) nogil

	# Calculate n hits within bins
	void ailist_bin_nhits(ailist_t *ail, long coverage[], int bin_size) nogil

	# Calculate n hits of a length within bins
	void ailist_bin_nhits_length(ailist_t *ail, long coverage[], int bin_size, int min_length, int max_length) nogil

	#-------------------------------------------------------------------------------------
	# ailist_wps.c
	#=====================================================================================

	# Calculate Window Protection Score
	void ailist_wps(ailist_t *ail, double wps[], uint32_t protection) nogil

	# Calculate Window Protection Score of a length
	void ailist_wps_length(ailist_t *ail, double wps[], uint32_t protection, int min_length, int max_length) nogil

	#-------------------------------------------------------------------------------------
	# ailist_merge.c
	#=====================================================================================

	# Merge nearby intervals
	ailist_t *ailist_merge(ailist_t *ail, uint32_t gap) nogil


	#-------------------------------------------------------------------------------------
	# ailist_extract.c
	#=====================================================================================

	# Extract start for ailist
	void ailist_extract_starts(ailist_t *ail, long starts[]) nogil

	# Extract end for ailist
	void ailist_extract_ends(ailist_t *ail, long ends[]) nogil

	# Extract index for ailist
	void ailist_extract_ids(ailist_t *ail, long ids[]) nogil


	#-------------------------------------------------------------------------------------
	# ailist_ops.c
	#=====================================================================================

	# Subtract intervals from region
	void ailist_subtract_intervals(interval_t *intv, ailist_t *ail, ailist_t *result_ail) nogil

	# Subtract two ailist_t intervals
	ailist_t *ailist_subtract(ailist_t *ref_ail, ailist_t *query_ail) nogil

	# Subtract intervals from region
	void ailist_common_intervals(interval_t *intv, ailist_t *ail, ailist_t *result_ail) nogil

	# Subtract two ailist_t intervals
	ailist_t *ailist_common(ailist_t *ail, ailist_t *other_ail) nogil

	# Union of two ailist_t intervals
	ailist_t *ailist_union(ailist_t *ail, ailist_t *other_ail) nogil


	#-------------------------------------------------------------------------------------
	# ailist_filter.c
	#=====================================================================================

	# Filter ailist by length
	void ailist_length_filter(ailist_t *ail, ailist_t *filtered_ail, int min_length, int max_length) nogil

	# Randomly downsample
	ailist_t *ailist_downsample(ailist_t *ail, double proportion) nogil


	#-------------------------------------------------------------------------------------
	# ailist_simulate.c
	#=====================================================================================

	# Simulate intervals
	void ailist_simulate(ailist_t *ail, ailist_t *simulation, int n) nogil


cpdef object rebuild_AIList(bytes data, bytes b_length)


cdef class AIList(object):
	"""
	Wrapper for C ailist_t
	"""

	# AIList attributes
	cdef ailist_t *c_ailist
	cdef public bint is_constructed
	cdef public bint is_closed
	cdef public bint is_frozen

	# Methods for serialization
	cdef bytes _get_data(self)
	cdef ailist_t *_set_data(self, bytes data, bytes b_length)

	# AIList methods
	cdef void set_list(AIList self, ailist_t *input_list)
	cdef void _insert(AIList self, int start, int end, int id_value)
	cdef void _construct(AIList self, int min_length)
	cdef ailist_t *_array_id(AIList self, const long[::1] ids)
	cdef ailist_t *_interval_id(AIList self, int id_value)
	cdef ailist_t *_intersect(AIList self, int start, int end)
	cdef ailist_t *_intersect_from_array(AIList self, const long[::1] starts, const long[::1] ends)
	cdef np.ndarray _intersect_ids(AIList self, int start, int end)
	cpdef _intersect_ids_from_array(AIList self, const long[::1] starts, const long[::1] ends, const long[::1] ids)
	cdef ailist_t *_intersect_from_ailist(AIList self, AIList ail)
	cdef np.ndarray _coverage(AIList self)
	cdef np.ndarray _bin_coverage(AIList self, int bin_size)
	cdef np.ndarray _bin_coverage_length(AIList self, int bin_size, int min_length, int max_length)
	cdef np.ndarray _bin_nhits(AIList self, int bin_size)
	cdef np.ndarray _bin_nhits_length(AIList self, int bin_size, int min_length, int max_length)
	cdef void _nhits(AIList self, long *nhits, int start, int end)
	cdef void _nhits_length(AIList self, long *nhits, int start, int end, int min_length, int max_length)
	cdef np.ndarray _wps(AIList self, int protection)
	cdef np.ndarray _wps_length(AIList self, int protection, int min_length, int max_length)
	cdef np.ndarray _length_dist(AIList self)
	cdef np.ndarray _nhits_from_array(AIList self, const long[::1] starts, const long[::1] ends)
	cdef np.ndarray _nhits_from_array_length(AIList self, const long[::1] starts, const long[::1] ends, int min_length, int max_length)
	cdef np.ndarray _interval_coverage(AIList self, int start, int end)


#cython: embedsignature=True
#cython: profile=False

import os
import sys
import numpy as np
cimport numpy as np
import math
cimport cython
import pandas as pd
from libc.string cimport memcpy
from .Interval_core import Interval
np.import_array()

# Set byteorder for __reduce__
byteorder = sys.byteorder


def get_include():
	"""
	Get file directory if C headers

	Parameters
	----------
		None

	Returns
	-------
		str (Directory to header files)
	"""

	return os.path.split(os.path.realpath(__file__))[0]


cpdef AIList rebuild_AIList(bytes data, bytes b_length):
	"""
	Rebuild function for __reduce__()

	Parameters
	----------
		data : bytes
			Bytes representation of ailist_t
		b_length : bytes
			Length of ailist_t

	Returns
	-------
		c : ailist_t*
			Translated ailist_t from data
	"""

	# Initialize new AIList
	c = AIList()

	# Build ailist from serialized data
	cdef ailist_t *interval_list = c._set_data(data, b_length)
	c.set_list(interval_list)

	return c


@cython.auto_pickle(True)
cdef class AIList(object):
	"""
	Wrapper for C ailist_t

	:class:`~ailist.AIList` stores a list of intervals
	"""

	def __cinit__(self):
		"""
		Initialize AIList object
		"""

		self.c_ailist = ailist_init()
		self.is_constructed = False
		self.is_closed = False
		self.is_frozen = False


	def __init__(self):
		"""
		Initialize AIList object

		Parameters
		----------
			None

		Returns
		-------
			None

		"""

		pass


	def __dealloc__(self):
		"""
		Free AIList.c_ailist
		"""

		ailist_destroy(self.c_ailist)


	cdef bytes _get_data(self):
		"""
		Function to convert ailist_t to bytes
		for serialization by __reduce__()
		"""

		return <bytes>(<char*>self.c_ailist.interval_list)[:(sizeof(interval_t)*self.c_ailist.nr)]


	cdef ailist_t *_set_data(self, bytes data, bytes b_length):
		"""
		Function to build ailist_t object from
		serialized bytes using __reduce__()

		Parameters
		----------
			data : bytes
				Bytes representation of ailist_t
			b_length : bytes
				Length of ailist_t

		Returns
		-------
			interval_list : ailist_t*
				Translated ailist_t for bytes
		"""

		# Convert bytes to ints
		cdef int length = int.from_bytes(b_length, byteorder)

		# Create new ailist_t
		cdef ailist_t *ail = ailist_init()
		cdef interval_t *interval_list = <interval_t*>malloc(length * sizeof(interval_t))
		memcpy(interval_list, <char*>data, sizeof(interval_t)*length)

		# Iteratively add intervals to interval_list
		cdef int i
		for i in range(length):
			ailist_add(ail, interval_list[i].start, interval_list[i].end,
						interval_list[i].id_value)

		return ail


	def __reduce__(self):
		"""
		Used for pickling. Convert ailist to bytes and back.
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Convert ints to bytes
		b_length = int(self.c_ailist.nr).to_bytes(4, byteorder)

		# Convert ailist_t to bytes
		data = self._get_data()

		return (rebuild_AIList, (data, b_length))


	@property
	def nc(self):
		"""
		Number of components in constructed AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if object is constructed
		if self.is_constructed:
			return self.c_ailist.nc
		else:
			return None


	@property
	def lenC(self):
		"""
		Length of components in constructed AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if object is constructed
		if self.is_constructed:
			return self.c_ailist.lenC[self.c_ailist.nc]
		else:
			return None


	@property
	def idxC(self):
		"""
		Index of components in constructed AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if object is constructed
		if self.is_constructed:
			return self.c_ailist.idxC[self.c_ailist.nc]
		else:
			return None


	@property
	def size(self):
		"""
		Number of intervals in AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		return self.c_ailist.nr

	@property
	def first(self):
		"""
		Start of first interval in AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if there are any intervals
		if self.size == 0:
			return None
		else:
			return self.c_ailist.first

	@property
	def last(self):
		"""
		End of last intervals in AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if there are any intervals
		if self.size == 0:
			return None
		else:
			return self.c_ailist.last

	@property
	def range(self):
		"""
		AIList.last - AIList.first
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if ther are any intervals
		if self.size == 0:
			return 0
		else:
			return self.last - self.first


	def __len__(self):
		"""
		Return size of interval_list
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		return self.size


	def __iter__(self):
		"""
		Iterate over AIList object
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Iterate over interval list
		cdef int i
		for i in range(self.size):
			interval = Interval(self.c_ailist.interval_list[i].start,
								self.c_ailist.interval_list[i].end)

			yield interval


	def __sub__(self, AIList query_ail):
		"""
		Subtract values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		return self.subtract(query_ail)


	def __add__(self, AIList query_ail):
		"""
		Common values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		return self.common(query_ail)


	def __or__(self, AIList query_ail):
		"""
		Common values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		return self.append(query_ail)


	def __hash__(self):
		"""
		Get hash value
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		return hash(self)


	cdef ailist_t *_array_id(AIList self, const long[::1] ids):
		cdef int length = len(ids)
		cdef ailist_t *cindexed_ailist

		# Call C function
		cindexed_ailist = ailist_get_id_array(self.c_ailist, &ids[0], length)

		return cindexed_ailist

	cdef ailist_t *_interval_id(AIList self, int id_value):
		cdef ailist_t *cindexed_ailist

		# Call C function
		cindexed_ailist = ailist_get_id(self.c_ailist, id_value)

		return cindexed_ailist

	def __getitem__(self, key):
		"""
		Index Intervals by id_value
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize variables
		cdef AIList indexed_ailist
		cdef ailist_t *cindexed_ailist
		cdef interval_t cinterval

		# Initialize results
		indexed_ailist = AIList()

		# Check if key is iterable
		try:
			# Test if iterable
			iter(key)

			# Check if keys are booleans
			if isinstance(key[0], np.bool_):
				raise IndexError("Cannot use boolean array as key")

			# Must be integers
			else:
				cindexed_ailist = self._array_id(key)

		# key is not iterable, treat as int
		except TypeError:
			# Check if key is slice
			if isinstance(key, slice):
				raise IndexError("Cannot use slice as key")

			# Find id
			#cindexed_ailist = self._interval_id(key)
			cinterval = self.c_ailist.interval_list[key]
			interval = Interval(cinterval.start, cinterval.end)
			#interval.set_i(cinterval)

			return interval

		# Wrap c object
		indexed_ailist.set_list(cindexed_ailist)

		return indexed_ailist


	def __repr__(self):
		"""
		Representation of ailist object
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize string
		repr_string = "AIList\n"
		if self.size == 0:
			repr_string += " range: (None-None)\n"
		else:
			repr_string += " range: (%d-%d)\n" % (self.first, self.last)

		# Iterate over interval_list
		if self.c_ailist.nr > 10:
			for i in range(5):
				repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)
			repr_string += "   ...\n"
			for i in range(-5, 0, 1):
				repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)
		else:
			for i in range(self.c_ailist.nr):
				repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)

		return repr_string


	cdef void set_list(AIList self, ailist_t *input_list):
		"""
		Set wrapper of C ailist

		Parameters
		----------
			input_list : ailist_t*
				ailist_t to replace existing one

		Returns
		-------
			None
		"""

		ailist_destroy(self.c_ailist)

		# Replace new skiplist
		self.c_ailist = input_list
		self.is_closed = False


	@property
	def ids(self):
		"""
		Return the ID values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Extract id_values
		cdef long[::1] ids = np.zeros(self.size, dtype=np.int_)
		ailist_extract_ids(self.c_ailist, &ids[0])

		return np.asarray(ids, dtype=np.intc)


	@property
	def starts(self):
		"""
		Return the start values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Extract start values
		cdef long[::1] starts = np.zeros(self.size, dtype=np.int_)
		ailist_extract_starts(self.c_ailist, &starts[0])

		return np.asarray(starts, dtype=np.intc)


	@property
	def ends(self):
		"""
		Return the end values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Extract end values
		cdef long[::1] ends = np.zeros(self.size, dtype=np.int_)
		ailist_extract_ends(self.c_ailist, &ends[0])

		return np.asarray(ends, dtype=np.intc)


	def freeze(self):
		"""
		Make :class:`~ailist.AIList` immutable

		Parameters
		----------
			None

		Returns
		-------
			None

		See Also
		--------
		AIList.unfreeze: Make mutable
		AIList.sort: Sort intervals inplace
		AIList.construct: Construct AIList, required to call AIList.intersect

		Examples
		--------
		>>> from ailist import AIList
		>>> ail = AIList()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(3, 6)
		>>> ail
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (3-6, 2)
		>>> ail.freeze()
		>>> ail.add(9, 10)
		TypeError: AIList is frozen and currently immutatable. Try '.unfreeze()' to reverse.

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure it is constructed
		if self.is_constructed is False:
			self.construct()

		# Change to frozen
		self.is_frozen = True


	def unfreeze(self):
		"""
		Make :class:`~ailist.AIList` mutable

		Parameters
		----------
			None

		Returns
		-------
			None

		See Also
		--------
		AIList.freeze: Make immutable
		AIList.sort: Sort intervals inplace
		AIList.construct: Construct AIList, required to call AIList.intersect

		Examples
		--------
		>>> from ailist import AIList
		>>> ail = AIList()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(3, 6)
		>>> ail
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (3-6, 2)
		>>> ail.freeze()
		>>> ail.add(9, 10)
		TypeError: AIList is frozen and currently immutatable. Try '.unfreeze()' to reverse.
		>>> ail.unfreeze()
		>>> ail.add(9, 10)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Change to not frozen
		self.is_frozen = False


	cdef void _insert(AIList self, int start, int end, int id_value):
		ailist_add(self.c_ailist, start, end, id_value)

	def add(self, start, end, id_value=None):
		"""
		Add an interval to AIList inplace

		Parameters
		----------
			start : int
				Start position of interval
			end : int
				End position of interval
			id_value : double
				ID of interval [default = None]

		Returns
		-------
			None

		See Also
		--------
		AIList.from_array: Add intervals from arrays
		AIList.construct: Construct AIList, required to call AIList.intersect
		AIList.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from ailist import AIList
		>>> ail = AIList()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(3, 6)
		>>> ail
		AIList
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (3-6)
		>>> import numpy as np
		>>> starts = np.arange(100)
		>>> ends = starts + 10
		>>> index = np.arange(len(starts))
		>>> ail = AIList()
		>>> ail.add(starts, ends, index)
		>>> ail
		AIList
		 range: (0-109)
			(0-10)
			(1-11)
			(2-12)
			(3-13)
			(4-14)
			...
			(95-105)
			(96-106)
			(97-107)
			(98-108)
			(99-109)


		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check that object is not frozen
		if self.is_frozen:
			raise TypeError("AIList is frozen and currently immutatable. Try '.unfreeze()' to reverse.")

		# Check if start is int
		cdef int array_length
		cdef const long[::1] starts
		cdef const long[::1] ends
		cdef const long[::1] ids
		if isinstance(start, int):
			# Check interval
			if start > end:
				raise IndexError("Start is greater than end.")

			# Insert interval
			if id_value is None:
				self._insert(start, end, self.c_ailist.nr)
			else:
				self._insert(start, end, id_value)

		elif isinstance(start, np.ndarray):
			array_length = len(start)
			starts = start
			ends = end
			ids = id_value
			ailist_from_array(self.c_ailist, &starts[0], &ends[0], &ids[0], array_length)

		else:
			raise TypeError("Start must be int or np.ndarray.")

		# Object is no longer constructed
		self.is_constructed = False

	@staticmethod
	def from_array(const long[::1] starts, const long[::1] ends, const long[::1] ids):
		"""
		Add intervals from arrays to AIList inplace

		Parameters
		----------
			starts : ~numpy.ndarray{long}
				Start positions of intervals
			ends : numpy.ndarray{long}
				End positions of intervals
			ids : numpy.ndarray{long}
				ID of intervals

		Returns
		-------
			None

		See Also
		--------
		AIList.add: Add interval to AIList
		AIList.construct: Construct AIList, required to call AIList.intersect
		AIList.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from ailist import AIList
		>>> import numpy as np
		>>> starts = np.arange(100)
		>>> ends = starts + 10
		>>> index = np.arange(len(starts))
		>>> ail = AIList.from_array(starts, ends, index)
		>>> ail
		AIList
		  range: (0-109)
		   (0-10, 0)
		   (1-11, 1)
		   (2-12, 2)
		   (3-13, 3)
		   (4-14, 4)
		   ...
		   (95-105, 95)
		   (96-106, 96)
		   (97-107, 97)
		   (98-108, 98)

		"""

		ail = AIList()
		cdef int array_length = len(starts)
		ailist_from_array(ail.c_ailist, &starts[0], &ends[0], &ids[0], array_length)

		return ail


	cdef void _construct(AIList self, int min_length):
		# Contruct
		ailist_construct(self.c_ailist, min_length)
		#ailist_construct_v0(self.c_ailist, min_length)

	def construct(self, int min_length=20):
		"""
		Construct ailist_t *Required to call intersect

		Parameters
		----------
			min_length : int
				Minimum length

		Returns
		-------
			None

		.. warning::
			This will re-sort intervals inplace

		See Also
		--------
		AIList.sort: Sort intervals inplace
		AIList.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from ailist import AIList
		>>> ail = AIList()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(2, 6)
		>>> ail
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (2-6, 2)
		>>> ail.construct()
		>>> ail
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (2-6, 2)
		   (3-4, 1)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if already constructed
		if self.is_constructed is False:
			self._construct(min_length)
			self.is_constructed = True
		else:
			pass


	def iter_sorted(self):
		"""
		Iterate over an AIList in sorted way

		Parameters
		----------
			None

		Returns
		-------
			sorted_iter : Generator
				Generator of Intervals
		"""

		# Check if is constructed
		if self.is_constructed is False:
			self.construct()

		# Iterate over  ail
		cdef interval_t *cintv
		#cdef Interval output_interval

		# Create sorted iterators
		cdef ailist_sorted_iter_t *ail_iter = ailist_sorted_iter_init(self.c_ailist)
		while ailist_sorted_iter_next(ail_iter) != 0:
			cintv = ail_iter.intv
			# Create Interval wrapper
			output_interval = Interval(cintv.start, cintv.end)
			#output_interval.set_i(cintv)
			yield output_interval

		ailist_sorted_iter_destroy(ail_iter)


	cdef ailist_t *_intersect(AIList self, int start, int end):
		cdef ailist_t *overlaps = ailist_init()
		ailist_query(self.c_ailist, overlaps, start, end)

		return overlaps

	cdef ailist_t *_intersect_from_array(AIList self,
										 const long[::1] starts,
										 const long[::1] ends):
		cdef int length = starts.size
		cdef ailist_t *overlaps = ailist_init()
		ailist_query_from_array(self.c_ailist, overlaps, &starts[0], &ends[0], length)

		return overlaps

	def intersect(self, start, end):
		"""
		Find intervals overlapping given range

		Parameters
		----------
			start : int | np.ndarray{int}
				Start position of query range
			end : int | np.ndarray{int}
				End position of query range

		Returns
		-------
			overlaps : AIList
				Overlapping intervals

		.. warning::
			This requires :func:`~ailist.AIList.construct` and will run it if not already run.
			This will re-sort intervals inplace.

		See Also
		--------
		AIList.construct: Construct AIList, required to call AIList.intersect
		AIList.add: Add interval to AIList
		AIList.intersect_index: Find interval indices overlapping given range
		AIList.intersect_from_array: Find interval indices overlapping given ranges

		Examples
		--------
		>>> from ailist import AIList
		>>> ail = AIList()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(2, 6)
		>>> ail
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (2-6, 2)
		>>> q = ail.intersect(2, 10)
		>>> q
		AIList
		  range: (2-6)
		   (2-6, 2)
		   (3-4, 1)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if is constructed
		if self.is_constructed is False:
			self.construct()

		# Intersect
		cdef const long[::1] starts
		cdef const long[::1] ends
		cdef ailist_t *i_list
		cdef AIList overlaps = AIList()
		if isinstance(start, int):
			i_list = self._intersect(start, end)
			overlaps.set_list(i_list)
		elif isinstance(start, np.ndarray):
			starts = start
			ends = end
			i_list = self._intersect_from_array(starts, ends)
			overlaps.set_list(i_list)
		else:
			raise TypeError("Start must be int or np.ndarray.")

		return overlaps


	cdef np.ndarray _intersect_ids(AIList self, int start, int end):
		cdef ailist_t *overlaps = ailist_init()
		ailist_query(self.c_ailist, overlaps, start, end)
		cdef long[::1] ids = np.zeros(overlaps.nr, dtype=np.long)

		# Extract IDs
		ailist_extract_ids(overlaps, &ids[0])

		return np.asarray(ids)

	@cython.boundscheck(False)
	@cython.wraparound(False)
	@cython.initializedcheck(False)
	cpdef _intersect_ids_from_array(AIList self, const long[::1] starts, const long[::1] ends,
									const long[::1] ids):
		cdef int length = len(starts)
		cdef array_query_t *total_overlaps = array_query_init()
		ailist_query_id_from_array(self.c_ailist, total_overlaps, &starts[0],
									&ends[0], &ids[0], length)

		cdef np.ndarray ref_index = pointer_to_numpy_array(total_overlaps.ref_index,
														   total_overlaps.size)
		cdef np.ndarray query_index = pointer_to_numpy_array(total_overlaps.query_index,
															 total_overlaps.size)

		return ref_index, query_index

	def intersect_ids(self, start, end):
		"""
		Find interval indices overlapping given range

		Parameters
		----------
			start : int | np.ndarray{int}
				Start position of query range
			end : int | np.ndarray{int}
				End position of query range

		Returns
		-------
			ids : numpy.ndarray{int}
				Overlapping interval indices

		.. warning::
			This requires :func:`~ailist.AIList.construct` and will run it if not already run.
			This will re-sort intervals inplace.

		See Also
		--------
		AIList.construct: Construct AIList, required to call AIList.intersect
		AIList.add: Add interval to AIList
		AIList.intersect: Find intervals overlapping given range
		AIList.intersect_from_array: Find interval indices overlapping given ranges

		Examples
		--------
		>>> from ailist import AIList
		>>> ail = AIList()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(2, 6)
		>>> ail
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (2-6, 2)
		>>> q = ail.intersect_index(2, 10)
		>>> q
		array([2, 1])

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if object has been constructed
		if self.is_constructed is False:
			self.construct()

		# Intersect
		cdef const long[::1] starts
		cdef const long[::1] ends
		cdef np.ndarray ref_ids
		cdef np.ndarray query_ids
		if isinstance(start, int):
			ref_ids = self._intersect_ids(start, end)
			return ref_ids

		elif isinstance(start, np.ndarray):
			starts = start
			ends = end
			ids = np.arange(len(starts), dtype=np.long)
			ref_ids, query_ids = self._intersect_ids_from_array(starts, ends, ids)
			return ref_ids, query_ids

		else:
			raise TypeError("Start must be int or np.ndarray.")


	cdef ailist_t *_intersect_from_ailist(AIList self, AIList ail):
		# Intersect with other AIList
		cdef ailist_t *overlaps = ailist_init()
		ailist_query_from_ailist(self.c_ailist, ail.c_ailist, overlaps)

		return overlaps

	def intersect_from_ailist(self, AIList ail_query):
		"""
		Find interval indices overlapping given ranges

		Parameters
		----------
			ail_query : AIList
				Intervals to query

		Returns
		-------
			ail : AIList
				Overlapping intervals

		.. warning::
			This requires :func:`~ailist.AIList.construct` and will run it if not already run.
			This will re-sort intervals inplace.

		See Also
		--------
		AIList.construct: Construct AIList, required to call AIList.intersect
		AIList.add: Add interval to AIList
		AIList.intersect: Find intervals overlapping given range
		AIList.intersect_index: Find interval indices overlapping given range
		AIList.intersect_from_array: Find interval indices overlapping given range

		Examples
		--------
		>>> from ailist import AIList
		>>> ail1 = AIList()
		>>> ail1.add(1, 2)
		>>> ail1.add(3, 4)
		>>> ail1.add(2, 6)
		>>> ail1
		AIList
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (2-6, 2)
		>>> ail2 = AIList()
		>>> ail2.add(1, 2)
		>>> ail2.add(3, 6)
		>>> ail2
		AIList
		  range: (1-6)
		    (1-2, 0)
		    (3-6, 1)
		>>> q = ail1.intersect_from_ailist(ail2)
		>>> q
		(array([0, 1, 1]), array([0, 2, 1]))

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Check if object is constructed
		if self.is_constructed is False:
			self.construct()

		# Intersect
		cdef ailist_t *c_ailist = self._intersect_from_ailist(ail_query)
		cdef AIList ail = AIList()
		ail.set_list(c_ailist)

		return ail


	cdef np.ndarray _coverage(AIList self):
		# Initialize coverage
		cdef double[::1] coverage = np.zeros(self.range, dtype=np.double)

		# Call C function
		ailist_coverage(self.c_ailist, &coverage[0])

		return np.asarray(coverage)

	def coverage(self):
		"""
		Find number of intervals overlapping every
		position in the AList range

		Parameters
		----------
			None

		Returns
		-------
			coverage : pandas.Series{double}
				Position on index and coverage as values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize coverage
		cdef np.ndarray coverage

		# Calculate coverage
		coverage = self._coverage()

		return pd.Series(coverage, index=np.arange(self.first, self.last))


	cdef np.ndarray _bin_coverage(AIList self, int bin_size):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef double[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		ailist_bin_coverage(self.c_ailist, &bins[0], bin_size)

		return np.asarray(bins)

	cdef np.ndarray _bin_coverage_length(AIList self, int bin_size, int min_length,
										 int max_length):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef double[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		ailist_bin_coverage_length(self.c_ailist, &bins[0], bin_size, min_length, max_length)

		return np.asarray(bins)

	def bin_coverage(self, int bin_size=100000, min_length=None, max_length=None):
		"""
		Find sum of coverage within binned
		positions

		Parameters
		----------
			bin_size : int
				Size of the bin to use
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			bins : pandas.Series{double}
				Position on index and coverage as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize coverage
		cdef np.ndarray bins

		# Calculate coverage
		if min_length is None or max_length is None:
			bins = self._bin_coverage(bin_size)
		else:
			bins = self._bin_coverage_length(bin_size, min_length, max_length)

		return pd.Series(bins, index=(np.arange(len(bins)) + int(self.first / bin_size)) * bin_size)


	cdef np.ndarray _bin_nhits(AIList self, int bin_size):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef long[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		ailist_bin_nhits(self.c_ailist, &bins[0], bin_size)

		return np.asarray(bins)

	cdef np.ndarray _bin_nhits_length(AIList self, int bin_size, int min_length, int max_length):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef long[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		ailist_bin_nhits_length(self.c_ailist, &bins[0], bin_size, min_length, max_length)

		return np.asarray(bins)

	def bin_nhits(self, int bin_size=100000, min_length=None, max_length=None):
		"""
		Find number of intervals overlapping binned
		positions

		Parameters
		----------
			bin_size : int
				Size of the bin to use
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			bins : pandas.Series{double}
				Position on index and coverage as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize coverage
		cdef np.ndarray bins

		# Calculate coverage
		if min_length is None or max_length is None:
			bins = self._bin_nhits(bin_size)
		else:
			bins = self._bin_nhits_length(bin_size, min_length, max_length)

		return pd.Series(bins, index=(np.arange(len(bins)) + int(self.first / bin_size)) * bin_size)


	cdef void _nhits(AIList self, long *nhits, int start, int end):

		# Call C function
		ailist_query_nhits(self.c_ailist, nhits, start, end)

		return

	cdef void _nhits_length(AIList self, long *nhits, int start, int end, int min_length,
							int max_length):

		# Call C function
		ailist_query_nhits_length(self.c_ailist, nhits, start, end, min_length, max_length)

		return

	def nhits(self, start, end, min_length=None, max_length=None):
		"""
		Find number of intervals overlapping binned
		positions

		Parameters
		----------
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			nhits : int
				Number of overlaps

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize nhits
		cdef long nhits = 0

		# Calculate coverage
		if min_length is None or max_length is None:
			bins = self._nhits(&nhits, start, end)
		else:
			bins = self._nhits_length(&nhits, start, end, min_length, max_length)

		return nhits


	def display(self):
		"""
		Print all intervals
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		display_list(self.c_ailist)


	def merge(self, int gap=0):
		"""
		Merge intervals within a gap

		Parameters
		----------
			gap : int
				Gap between intervals to merge

		Returns
		-------
			merged_list : AIList
				Merged intervals

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure list is constructed
		if self.is_constructed is False:
			self.construct()

		# Create merged
		cdef AIList merged_list = AIList()

		# Call C function
		cdef ailist_t *merged_clist = ailist_merge(self.c_ailist, gap)
		merged_list.set_list(merged_clist)

		return merged_list


	def subtract(self, AIList query_ail):
		"""
		Subtract intervals within another AIList

		Parameters
		----------
			query_ail : AIList
				AIList of intervals to subtract

		Returns
		-------
			subtracted_list : AIList
				Subtracted intervals
		"""

		# Check if object is still open
		if self.is_closed or query_ail.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure list is constructed
		if self.is_constructed is False:
			self.construct()
		if query_ail.is_constructed is False:
			query_ail.construct()

		# Create subracted
		cdef AIList subtracted_list = AIList()

		# Call Cfunction
		cdef ailist_t *subtracted_clist = ailist_subtract(self.c_ailist,
														  query_ail.c_ailist)
		subtracted_list.set_list(subtracted_clist)

		return subtracted_list


	def common(self, AIList query_ail):
		"""
		Common intervals within another AIList

		Parameters
		----------
			query_ail : AIList
				AIList of intervals to find commons

		Returns
		-------
			common_list : AIList
				Common intervals

		"""

		# Check if object is still open
		if self.is_closed or query_ail.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure list is constructed
		if self.is_constructed is False:
			self.construct()
		if query_ail.is_constructed is False:
			query_ail.construct()

		# Create common
		cdef AIList common_list = AIList()
		cdef ailist_t *common_clist = ailist_common(self.c_ailist,
													query_ail.c_ailist)
		common_list.set_list(common_clist)

		return common_list


	def append(self, AIList query_ail):
		"""
		Union of intervals within two AIList

		Parameters
		----------
			query_ail : AIList
				AIList of intervals to append

		Returns
		-------
			union_list: AIList
				Union of intervals

		"""

		# Check if object is still open
		if self.is_closed or query_ail.is_closed:
			raise NameError("AIList object has been closed.")

		# Create union
		cdef AIList union_list = AIList()
		cdef ailist_t *union_clist = ailist_append(query_ail.c_ailist,
												   self.c_ailist)
		union_list.set_list(union_clist)

		return union_list


	cdef np.ndarray _wps(AIList self, int protection):
		# Initialize wps
		cdef double[::1] wps = np.zeros(self.range, dtype=np.double)

		ailist_wps(self.c_ailist, &wps[0], protection)

		return np.asarray(wps)

	cdef np.ndarray _wps_length(AIList self, int protection, int min_length, int max_length):
		# Initialize wps
		cdef double[::1] wps = np.zeros(self.range, dtype=np.double)

		ailist_wps_length(self.c_ailist, &wps[0], protection, min_length, max_length)

		return np.asarray(wps)

	def wps(self, int protection=60, min_length=None, max_length=None):
		"""
		Calculate Window Protection Score
		for each position in AIList range

		Parameters
		----------
			protection : int
				Protection window to use
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			wps : pandas.Series{double}
				Position on index and WPS as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure list is constructed
		if self.is_constructed is False:
			self.construct()

		# Initialize wps
		cdef np.ndarray wps

		# Calculate wps
		if self.range == 0:
			return None
		if min_length is None or max_length is None:
			wps = self._wps(protection)
		else:
			wps = self._wps_length(protection, min_length, max_length)

		return pd.Series(wps, index=np.arange(self.first, self.last))


	def filter(self, int min_length=1, int max_length=400):
		"""
		Filter out intervals outside of a length range

		Parameters
		----------
			min_length : int
				Minimum length to keep
			max_length : int
				Maximum langth to keep

		Returns
		-------
			filtered_ail : AIList
				Filtered intervals

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize filtered list
		cdef AIList filtered_ail = AIList()

		# Call C function
		ailist_length_filter(self.c_ailist, filtered_ail.c_ailist, min_length, max_length)
		#filtered_ail.set_list(cfiltered_ail)

		return filtered_ail


	cdef np.ndarray _length_dist(AIList self):
		# Initialize distribution
		cdef int max_length = ailist_max_length(self.c_ailist)
		cdef int[::1] distribution = np.zeros(max_length + 1, dtype=np.intc)

		# Calculate distribution
		ailist_length_distribution(self.c_ailist, &distribution[0])

		return np.asarray(distribution, dtype=np.intc)

	def length_dist(self):
		"""
		Calculate length distribution of intervals

		Parameters
		----------
			None

		Returns
		-------
			distribution : numpy.ndarray{int}
				Interval length distribution

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize distribution
		cdef np.ndarray distribution

		# Calculate distribution
		distribution = self._length_dist()

		return distribution


	cdef np.ndarray _nhits_from_array(AIList self, const long[::1] starts, const long[::1] ends):
		# Initialize hits
		cdef int length = starts.size
		cdef int[::1] nhits = np.zeros(length, dtype=np.intc)

		# Calculate distribution
		ailist_nhits_from_array(self.c_ailist, &starts[0], &ends[0], length, &nhits[0])

		return np.asarray(nhits, dtype=np.intc)

	cdef np.ndarray _nhits_from_array_length(AIList self, const long[::1] starts,
											 const long[::1] ends, int min_length,
											 int max_length):
		# Initialize hits
		cdef int length = starts.size
		cdef int[::1] nhits = np.zeros(length, dtype=np.intc)

		# Calculate distribution
		ailist_nhits_from_array_length(self.c_ailist, &starts[0], &ends[0], length,
										&nhits[0], min_length, max_length)

		return np.asarray(nhits, dtype=np.intc)

	def nhits_from_array(self, const long[::1] starts, const long[::1] ends,
						 min_length=None, max_length=None):
		"""
		Find number of intervals overlapping given
		positions

		Parameters
		----------
			starts : numpy.ndarray{long}
				Start positions to intersect
			ends : numpy.ndarray{long}
				End positions to intersect
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			nhits : numpy.ndarray{int}
				Number of hits per position

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure list is constructed
		if self.is_constructed is False:
			self.construct()

		# Initialize distribution
		cdef np.ndarray nhits
		# Calculate distribution
		if min_length is None or max_length is None:
			nhits = self._nhits_from_array(starts, ends)
		else:
			nhits = self._nhits_from_array_length(starts, ends, min_length, max_length)

		return nhits


	cdef np.ndarray _interval_coverage(AIList self, int start, int end):
		# Initialize hits
		cdef int[::1] coverage = np.zeros(end - start, dtype=np.intc)

		# Calculate distribution
		ailist_interval_coverage(self.c_ailist, start, end, &coverage[0])

		return np.asarray(coverage, dtype=np.intc)

	def interval_coverage(self, int start, int end):
		"""
		Find number of intervals overlapping each
		position in given interval

		Parameters
		----------
			start : int
				Start position to intersect
			end : int
				End position to intersect

		Returns
		-------
			coverage : pandas.Series{int}
				Number of hits per position

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Make sure list is constructed
		if self.is_constructed is False:
			self.construct()

		# Initialize distribution
		cdef np.ndarray coverage

		# Calculate distribution
		coverage = self._interval_coverage(start, end)

		return pd.Series(coverage, index=np.arange(start, end))


	def downsample(self, double proportion):
		"""
		Randomly downsample AIList

		Parameters
		----------
			proportion : double
				Proportion of intervals to keep

		Returns
		-------
			filtered_ail : AIList
				Downsampled AIList

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		# Initialize filtered list
		cdef AIList filtered_ail = AIList()

		# Call C function
		cdef ailist_t *cfiltered_ail = ailist_downsample(self.c_ailist, proportion)
		filtered_ail.set_list(cfiltered_ail)

		return filtered_ail


	def copy(self):
		"""
		Make a copy of the AIList
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("AIList object has been closed.")

		cdef AIList new_ail = AIList()
		cdef ailist_t *c_new_ail = ailist_copy(self.c_ailist)
		new_ail.set_list(c_new_ail)

		return new_ail


	def close(self):
		"""
		Close object and clear memory
		"""

		ailist_destroy(self.c_ailist)
		self.c_ailist = NULL

		self.is_closed = True



import numpy as np
cimport numpy as np
from libc.stdint cimport uint32_t, int32_t


cdef extern from "src/ailist/interval.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist/interval.h":
	# C is include here so that it doesn't need to be compiled externally
	ctypedef struct interval_t:
		uint32_t start      				# Region start: 0-based
		uint32_t end    					# Region end: not inclusive
		int32_t id_value					# Region ID

	# Initialize interval_t
	interval_t *interval_init(uint32_t start, uint32_t end, int32_t id_value) nogil


cdef class Interval(object):
	"""
	Wrapper for C interval
	"""

	# Interval attributes
	cdef public int start
	cdef public int end


from __future__ import absolute_import
from .AIList_core import AIList
from .IntervalArray_core import IntervalArray
from .Interval_core import Interval
from .LabeledIntervalArray_core import LabeledIntervalArray, LabeledInterval, get_include

# This is extracted automatically by the top-level setup.py.
__version__ = '2.1.0'
__author__ = "Kyle S. Smith"


__doc__ = """\
API
======

Basic class
-----------

.. autosummary::
   :toctree: .
   
   Interval
   LabeledInterval
   AIList
   IntervalArray
   LabeledIntervalArray
    
"""


import numpy as np
cimport numpy as np
from libc.stdint cimport uint32_t, int32_t, int64_t, int16_t
from .AIList_core cimport *
from .Interval_core cimport interval_init
from .array_query_core cimport *
ctypedef np.uint8_t uint8


cdef extern from "src/labeled_aiarray/labeled_augmented_array.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/overlap_label_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_add.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_get.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_construct.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_query_single.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_query_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_query_array.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_nhits.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_wps.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_coverage.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_merge.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_filter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_extract.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_iter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_match.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_sort.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_simulate.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_ops.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_aiarray_percent.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/khash.h":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray/labeled_augmented_array.h":

	ctypedef struct label_t:
		const char *name    				# Name of the label
		ailist_t *ail						# AIList object

	ctypedef struct labeled_interval_t:
		const char *name
		interval_t *i

	ctypedef struct labeled_aiarray_t:
		label_t *labels        				# List of Labels (of size _n_ctg_)
		int32_t n_labels, max_labels 		# Number and max number of labels
		void *label_lookup           	  	# Dict for converting label names to int
		uint32_t *first						# Record first position of intervals
		uint32_t *last						# Record last position of intervals
		int64_t total_nr					# Total number of regions
		uint32_t *id_index					# Record id index values
		int16_t is_constructed				# Flag for whether constructed or not

	ctypedef struct label_sorted_iter_t:
		const char *name					# Label
		ailist_sorted_iter_t *ail_iter		# Sorted AIList iterator
		interval_t *intv					# Interval

	ctypedef struct labeled_aiarray_iter_t:
		labeled_aiarray_t *laia				# Labeled aiarray
		int n								# Current position
		labeled_interval_t *intv			# Interval
		const char *name					# Label

	ctypedef struct labeled_aiarray_overlap_iter_t:
		labeled_aiarray_t *ref_laia		# Labeled aiarray
		labeled_aiarray_iter_t *query_iter	# Reference iterator
		labeled_aiarray_t *overlaps			# Overlaps

	ctypedef struct overlap_label_index_t:
		int size							# Current size
		int max_size						# Maximum size
		labeled_aiarray_t *laia				# Store labeled_aiarray
		long *indices						# Store indices


	#-------------------------------------------------------------------------------------
	# labeled_augmented_array.c
	#=====================================================================================

	# Initialize aiarray_t
	labeled_aiarray_t *labeled_aiarray_init() nogil

	# Free aiarray data
	void labeled_aiarray_destroy(labeled_aiarray_t *laia) nogil

	# Return index for given label
	int32_t get_label(const labeled_aiarray_t *laia, const char *label) nogil

	# Extract strings from array
	void slice_str(const char *str, char *buffer, size_t start, size_t end) nogil

	# Print labeled_aiarray
	void labeled_aiarray_print(labeled_aiarray_t *laia) nogil

	# Calculate maximum length
	int labeled_aiarray_max_length(labeled_aiarray_t *laia) nogil

	# Calculate length distribution
	void labeled_aiarray_length_distribution(labeled_aiarray_t *laia, int distribution[]) nogil

	#-------------------------------------------------------------------------------------
	# overlap_label_index.c
	#=====================================================================================

	# Initialize overlap_index
	overlap_label_index_t *overlap_label_index_init() nogil

	# Free overlap_label_index memory
	void overlap_label_index_destroy(overlap_label_index_t *oi) nogil

	# Add interval to overlap_label_index
	void overlap_label_index_add(overlap_label_index_t *oi, interval_t i, const char *label_name) nogil

	void overlap_label_index_wrap_ail(overlap_label_index_t * oi, ailist_t *ail, const char *label_name) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_add.c
	#=====================================================================================

	# Expand and add label
	void labeled_aiarray_add_label(labeled_aiarray_t *laia, const char*label) nogil

	# Add a interval_t interval
	void labeled_aiarray_add(labeled_aiarray_t *laia, uint32_t s, uint32_t e, const char *label) nogil

	# Link memory of single label labeled_aiarrays and transfer data owership
	void labeled_aiarray_multi_merge(labeled_aiarray_t *laia, labeled_aiarray_t **laia_array, int length) nogil

	# Add a interval_t from array
	void labeled_aiarray_from_array(labeled_aiarray_t *laia, const long starts[], const long ends[], const char label_names[], int length, int label_str_len) nogil

	# Add intervals from another labeled_aiarray
	void labeled_aiarray_append(labeled_aiarray_t *laia, labeled_aiarray_t *laia2) nogil

	# Copy labeled_aiarray
	labeled_aiarray_t *labeled_aiarray_copy(labeled_aiarray_t *laia) nogil

	# Append AIList in labeled_aiarry
	void labeled_aiarray_append_ail(labeled_aiarray_t *laia, ailist_t *ail, const char *label_name) nogil

	# Wrap AIList in labeled_aiarry
	void labeled_aiarray_wrap_ail(labeled_aiarray_t *laia, ailist_t *ail, const char *label_name) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_construct.c
	#=====================================================================================

	# Construct aiarray: decomposition and augmentation
	void labeled_aiarray_construct(labeled_aiarray_t *laia, int cLen) nogil

	# Validation construction ran
	int labeled_aiarray_validate_construction(labeled_aiarray_t *laia) nogil

	# Deconstruct
	void labeled_aiarray_deconstruct(labeled_aiarray_t *laia) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_query_single.c
	#=====================================================================================

	# Base query logic
	void labeled_aiarray_query(labeled_aiarray_t *laia, labeled_aiarray_t *overlaps, const char *label, uint32_t qs, uint32_t qe) nogil

	# Base query logic filtered by length
	void labeled_aiarray_query_length(labeled_aiarray_t *laia, labeled_aiarray_t *overlaps, const char *label_name, 
									uint32_t qs, uint32_t qe, int min_length, int max_length) nogil

	# Base query logic for nhits
	void labeled_aiarray_query_nhits(labeled_aiarray_t *laia, long *nhits, const char *label_name, uint32_t qs, uint32_t qe) nogil

	# Base query logic for nhits filtered by length
	void labeled_aiarray_query_nhits_length(labeled_aiarray_t *laia, long *nhits, const char *label_name, uint32_t qs,
											uint32_t qe, int min_length, int max_length) nogil

	# Base query logic if present
	void labeled_aiarray_query_has_hit(labeled_aiarray_t *laia, const char *label_name, uint8_t *has_hit, uint32_t qs, uint32_t qe) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_query_index.c
	#=====================================================================================

	# Base query logic with index
	void labeled_aiarray_query_with_index(labeled_aiarray_t *laia, const char *label_name, overlap_label_index_t *overlaps, uint32_t qs, uint32_t qe) nogil

	# Query with index from arrays
	void labeled_aiarray_query_with_index_from_array(labeled_aiarray_t *laia, overlap_label_index_t *overlaps, const char label_names[], const long starts[], const long ends[], int length, int label_str_len) nogil

	# Query with index from labeled_aiarray
	void labeled_aiarray_query_with_index_from_labeled_aiarray(labeled_aiarray_t *laia, labeled_aiarray_t *other_laia, overlap_label_index_t *overlaps) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_query_array.c
	#=====================================================================================

	# Base query logic with index
	void labeled_aiarray_query_only_index(labeled_aiarray_t *laia, const char *label_name, array_query_t *overlaps, uint32_t qs, uint32_t qe, uint32_t id) nogil

	# Query aiarray intervals from array
	array_query_t *labeled_aiarray_query_index_from_array(labeled_aiarray_t *laia, const char label_names[], const long starts[], const long ends[], int length, int label_str_len) nogil

	# Query intervals from array
	void labeled_aiarray_query_from_array(labeled_aiarray_t *laia, labeled_aiarray_t *overlaps, const char label_names[], const long starts[], const long ends[], int length, int label_str_len) nogil	

	# Query array if present
	void labeled_aiarray_query_has_hit_from_array(labeled_aiarray_t *laia, const char label_names[], const long starts[], const long ends[], int length, int label_str_len, uint8_t has_hit[]) nogil

	# Query aiarray intervals from aiarray
	void labeled_aiarray_query_from_labeled_aiarray(labeled_aiarray_t *laia, labeled_aiarray_t *laia2, labeled_aiarray_t *overlaps) nogil

	# Query aiarray intervals from aiarray
	void labeled_aiarray_query_index_from_labeled_aiarray(labeled_aiarray_t *laia, labeled_aiarray_t *laia2, array_query_t *aq) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_index.c
	#=====================================================================================

	void labeled_aiarray_cache_id(labeled_aiarray_t *ail) nogil

	labeled_interval_t *labeled_aiarray_get_index(labeled_aiarray_t *ail, int32_t i) nogil

	labeled_aiarray_t *labeled_aiarray_slice_index(labeled_aiarray_t *laia, const long ids[], int length) nogil

	labeled_aiarray_t *labeled_aiarray_slice_range(labeled_aiarray_t *laia, int start, int end, int step) nogil

	labeled_aiarray_t *labeled_aiarray_slice_bool(labeled_aiarray_t *laia, uint8_t bool_index[]) nogil

	int labeled_aiarray_index_with_aiarray(labeled_aiarray_t *laia, labeled_aiarray_t *other_laia) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_iter.c
	#=====================================================================================

	label_sorted_iter_t *label_sorted_iter_init(labeled_aiarray_t *laia, const char *label_name) nogil

	int label_sorted_iter_next(label_sorted_iter_t *iter) nogil

	void label_sorted_iter_destroy(label_sorted_iter_t *iter) nogil

	labeled_aiarray_iter_t *labeled_aiarray_iter_init(labeled_aiarray_t *laia) nogil

	int labeled_aiarray_iter_next(labeled_aiarray_iter_t *iter) nogil

	void labeled_aiarray_iter_destroy(labeled_aiarray_iter_t *iter) nogil

	labeled_aiarray_overlap_iter_t *labeled_aiarray_overlap_iter_init(labeled_aiarray_t *ref_laia, labeled_aiarray_t *query_laia) nogil

	int labeled_aiarray_overlap_iter_next(labeled_aiarray_overlap_iter_t *iter) nogil

	void labeled_aiarray_overlap_iter_destroy(labeled_aiarray_overlap_iter_t *iter) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_nhits.c
	#=====================================================================================

	void labeled_aiarray_nhits(labeled_aiarray_t *laia, long *nhits, const char *label_name, uint32_t qs, uint32_t qe) nogil

	void labeled_aiarray_nhits_length(labeled_aiarray_t *laia, long *nhits, const char *label_name, uint32_t qs,
											uint32_t qe, int min_length, int max_length) nogil

	void labeled_aiarray_nhits_from_array(labeled_aiarray_t *laia, const char label_names[], const long starts[], const long ends[],
											int length, int label_str_len, long *nhits) nogil

	void labeled_aiarray_nhits_from_array_length(labeled_aiarray_t *laia, const char label_names[], const long starts[], const long ends[],
											int length, int label_str_len, long *nhits, int min_length, int max_length) nogil

	void labeled_aiarray_nhits_from_labeled_aiarray(labeled_aiarray_t *laia, labeled_aiarray_t *other_laia,
													long *nhits) nogil

	void labeled_aiarray_nhits_from_labeled_aiarray_length(labeled_aiarray_t *laia, labeled_aiarray_t *other_laia,
													long *nhits, int min_length, int max_length) nogil

	void labeled_aiarray_has_hit(labeled_aiarray_t *laia, const char *label_name, uint8_t has_hit[], uint32_t qs, uint32_t qe) nogil

	void labeled_aiarray_bin_nhits(labeled_aiarray_t *laia, long *nhits, int bin_size) nogil

	void labeled_aiarray_bin_nhits_length(labeled_aiarray_t *laia, long *nhits, int bin_size, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_coverage.c
	#=====================================================================================

	# Determine coverage for an interval
	void labeled_aiarray_interval_coverage(labeled_aiarray_t *laia, int start, int end, const char *label_name, int coverage[]) nogil

	# Calculate coverage for a label
	void labeled_aiarray_label_coverage(labeled_aiarray_t *laia, double coverage[], const char *label_name) nogil

	# Calculate coverage for a label of a length
	void labeled_aiarray_label_coverage_length(labeled_aiarray_t *laia, double coverage[], const char *label_name,
												int min_length, int max_length) nogil

	# Calculate coverage within bins
	void labeled_aiarray_label_bin_coverage(labeled_aiarray_t *laia, double coverage[], int bin_size, const char *label_name) nogil

	# Calculate coverage of intervals of a given length within bins
	void labeled_aiarray_label_bin_coverage_length(labeled_aiarray_t *laia, double coverage[], int bin_size, const char *label_name,
													int min_length, int max_length) nogil
	
	# Calculate interval midpoint coverage
	void labeled_aiarray_label_midpoint_coverage(labeled_aiarray_t *laia, double coverage[], const char *label_name) nogil

	# Calculate interval midpoitn coverage with lengths
	void labeled_aiarray_label_midpoint_coverage_length(labeled_aiarray_t *laia, double coverage[], const char *label_name,
														int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_wps.c
	#=====================================================================================

	# Calculate Window Protection Score for a label
	void labeled_aiarray_label_wps(labeled_aiarray_t *laia, double wps[], uint32_t protection, const char *label_name) nogil

	# Calculate Window Protection Score for a label of a length
	void labeled_aiarray_label_wps_length(labeled_aiarray_t *laia, double wps[], uint32_t protection, int min_length, int max_length, const char *label_name) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_merge.c
	#=====================================================================================

	# Merge nearby intervals
	labeled_aiarray_t *labeled_aiarray_merge(labeled_aiarray_t *laia, uint32_t gap) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_filter.c
	#=====================================================================================

	# Filter labeled_aiarray by length 
	labeled_aiarray_t *labeled_aiarray_length_filter(labeled_aiarray_t *laia, int min_length, int max_length) nogil

	# Randomly downsample 
	labeled_aiarray_t *labeled_aiarray_downsample(labeled_aiarray_t *laia, double proportion) nogil

	overlap_label_index_t *labeled_aiarray_downsample_with_index(labeled_aiarray_t *laia, double proportion) nogil

	#-------------------------------------------------------------------------------------
	# labeled_aiarray_extract.c
	#=====================================================================================

	# Extract start for labeled_aiarray
	void labeled_aiarray_extract_starts(labeled_aiarray_t *laia, long starts[]) nogil

	# Extract end for labeled_aiarray
	void labeled_aiarray_extract_ends(labeled_aiarray_t *laia, long ends[]) nogil

	# Extract id for labeled_aiarray
	void labeled_aiarray_extract_ids(labeled_aiarray_t *laia, long ids[]) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_match.c
	#=====================================================================================

	# Return all exact matches between labeled_aiarrays
	overlap_label_index_t *labeled_aiarray_exact_match(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2) nogil

	void labeled_aiarray_has_exact_match(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2, uint8_t has_match[]) nogil

	# Return exact matches between labeled_aiarrays
	void labeled_aiarray_is_exact_match(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2, uint8_t has_match_laia1[], uint8_t has_match_laia2[]) nogil

	# Return exact matches between labeled_aiarrays
	void labeled_aiarray_which_exact_match(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2, array_query_t *matches) nogil

	# Return exact match between labeled_aiarray and an interval
	int labeled_aiarray_where_interval(labeled_aiarray_t *laia, const char *label, uint32_t qs, uint32_t qe) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_get.c
	#=====================================================================================

	# Get intervals with label name
	labeled_aiarray_t *labeled_aiarray_get_label(labeled_aiarray_t *laia, const char *label_name) nogil

	# Get intervals with label name
	labeled_aiarray_t *labeled_aiarray_view_label(labeled_aiarray_t *laia, const char *label_name) nogil

	# Get intervals with label name and original index
	overlap_label_index_t *labeled_aiarray_get_label_with_index(labeled_aiarray_t *laia, const char *label_name) nogil

	# Get intervals with labels names from array
	labeled_aiarray_t *labeled_aiarray_get_label_array(labeled_aiarray_t *laia, const char label_names[], int n_labels, int label_str_len) nogil

	# Get intervals with labels from array and original index
	overlap_label_index_t *labeled_aiarray_get_label_array_with_index(labeled_aiarray_t *laia, const char label_names[], int n_labels, int label_str_len) nogil

	# Determine if an index is of a label array
	void labeled_aiarray_get_label_array_presence(labeled_aiarray_t *ail, const char label_names[], int n_labels, uint8_t index[], int label_str_len) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_sort.c
	#=====================================================================================

	# Sort intervals by actual order
	void labeled_aiarray_order_sort(labeled_aiarray_t *laia) nogil

	# Sort intervals by starts
	void labeled_aiarray_sort_index(labeled_aiarray_t *laia, long *index) nogil

	# Sort intervals by starts
	void labeled_aiarray_sort(labeled_aiarray_t *laia) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_simulate.c
	#=====================================================================================

	# Simulate intervals
	void labeled_aiarray_simulate(labeled_aiarray_t *laia, labeled_aiarray_t *sim_laia) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_ops.c
	#=====================================================================================

	# Find common intervals between two labeled_aiarrays
	labeled_aiarray_t *labeled_aiarray_common(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2) nogil

	# Subtract intervals from laia1 that are in laia2
	labeled_aiarray_t *labeled_aiarray_subtract(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2) nogil

	# Find union of intervals between two labeled_aiarrays
	labeled_aiarray_t *labeled_aiarray_union(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_percent.c
	#=====================================================================================

	# Find percent coverage of intervals in laia1 that are in laia2
	void labeled_aiarray_percent_coverage(labeled_aiarray_t *laia1, labeled_aiarray_t *laia2, double coverage[]) nogil


#cpdef object rebuild_LabeledIntervalArray(bytes data, bytes b_length, bytes b_label_list)

cdef class LabeledInterval(object):
	"""
	Wrapper for C labeled interval
	"""

	# Define attributes
	cdef str _label

	# C instance of struct
	cdef interval_t *i

	# Methods for serialization
	cdef void set_i(self, interval_t *i, str label)


cdef class LabeledIntervalArray(object):
	"""
	Wrapper for C aiarray_t
	"""

	# Define class properties
	cdef labeled_aiarray_t *laia
	cdef public bint is_closed
	cdef public bint is_frozen

	# Functions for serialization
	#cdef bytes _get_data(self)
	#cdef labeled_aiarray_t *_set_data(self, bytes data, bytes b_length, bytes b_label_list)

	# Class methods
	cdef void set_list(LabeledIntervalArray self, labeled_aiarray_t *input_list)
	cdef ailist_t *_get_ail(LabeledIntervalArray self, char *label_name)

	cdef labeled_aiarray_t *_array_index(LabeledIntervalArray self, const long[::1] ids)
	cdef np.ndarray _get_index(LabeledIntervalArray self, str label)
	cdef np.ndarray _get_index_multi(LabeledIntervalArray self, np.ndarray labels)

	cdef void _insert(LabeledIntervalArray self, int start, int end, const char *label)
	cdef void _construct(LabeledIntervalArray self, int min_length)
	cdef void _from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char *label_names, int array_length, int label_str_len)
	cdef void _append(LabeledIntervalArray self, LabeledIntervalArray other_laia)

	#cdef int _get_label_index(LabeledIntervalArray self, int label)
	#cdef np.ndarray _get_label_comp_bounds(LabeledIntervalArray self, int label)
	#cdef np.ndarray _get_label_comp_length(LabeledIntervalArray self, int label)
	cdef labeled_aiarray_t *get_labels(LabeledIntervalArray self, const char *label_names, int str_label_len, int length)
	cdef overlap_label_index_t *get_labels_with_index(LabeledIntervalArray self, const char *label_names, int str_label_len, int length)

	cdef labeled_aiarray_t *_intersect(LabeledIntervalArray self, int start, int end, const char *label_name)
	cdef labeled_aiarray_t *_intersect_from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char *labels, int label_str_len)
	cpdef _intersect_from_array_only_index(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char *labels, int label_str_len)
	
	cdef void _has_hit_from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, uint8[::1] has_hit, const char *labels, int label_str_len)
	cdef labeled_aiarray_t *_intersect_from_labeled_aiarray(LabeledIntervalArray self, LabeledIntervalArray laia)
	cpdef _intersect_index_from_labeled_aiarray(LabeledIntervalArray self, LabeledIntervalArray laia)
	cpdef _intersect_with_index(LabeledIntervalArray self, int start, int end, const char *label_name)
	cpdef _intersect_with_index_from_labeled_aiarray(LabeledIntervalArray self, LabeledIntervalArray ail2)
	cpdef _intersect_from_array_with_index(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char *labels, int label_str_len)

	cdef void _nhits_from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, long[::1] nhits,
								const char *labels, int label_str_len)
	cdef void _nhits_from_array_length(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, long[::1] nhits,
										const char *labels, int label_str_len, int min_length, int max_length)

	cdef labeled_aiarray_t *_determine_bins(LabeledIntervalArray self, int bin_size)
	cdef void _bin_nhits(LabeledIntervalArray self, long[::1] nhits, int bin_size)
	cdef void _bin_nhits_length(LabeledIntervalArray self, long[::1] nhits, int bin_size, int min_length, int max_length)

	cdef np.ndarray _wps(LabeledIntervalArray self, const char *label_name, int protection)
	cdef np.ndarray _wps_length(LabeledIntervalArray self, const char *label_name, int protection, int min_length, int max_length)

	cdef np.ndarray _coverage(LabeledIntervalArray self, const char *label_name)
	cdef np.ndarray _coverage_length(LabeledIntervalArray self, const char *label_name, int min_length, int max_length)

	cpdef _downsample_with_index(LabeledIntervalArray self, double proportion)
	cdef np.ndarray _length_dist(LabeledIntervalArray self)
	cpdef _filter_exact_match(LabeledIntervalArray self, LabeledIntervalArray other_aiarray)
	cdef void _has_exact_match(LabeledIntervalArray self, LabeledIntervalArray other_aiarray, uint8_t[::1] has_match)
	cdef void _is_exact_match(LabeledIntervalArray self, LabeledIntervalArray other_aiarray, uint8_t[::1] has_match1, uint8_t[::1] has_match2)
	cpdef _which_exact_match(LabeledIntervalArray self, LabeledIntervalArray other_aiarray)
	cdef int _where_interval(LabeledIntervalArray self, int start, int end, const char *label_name)
	cdef int _index_with_aiarray(LabeledIntervalArray self, LabeledIntervalArray other_aiarray)
	cdef void _get_locs(LabeledIntervalArray self, uint8[::1] locs_view, char *label_names, int label_str_len, int n_labels)

	cdef void _create_bin(LabeledIntervalArray self, int bin_size, int bin_range, const char* label_name)
	cdef np.ndarray _midpoint_coverage(LabeledIntervalArray self, const char *label_name)

	cdef np.ndarray _percent_coverage(LabeledIntervalArray self, LabeledIntervalArray other_laia)




from .LabeledIntervalArray_core import LabeledIntervalArray, LabeledInterval
from .Interval_core import Interval
import numpy as np


class IntervalArray(object):
    """
    :class:`~IntervalArray.IntervalArray` stores a list of intervals
    """

    def __init__(self):
        """
        Initialize IntervalArray object

        Parameters
        ----------
            None

        Returns
        -------
            None

        """

        self._laia = LabeledIntervalArray()

    
    @property
    def is_closed(self):
        """
        Whether object is still in memory
        """

        self._laia.is_closed

    @property
    def is_frozen(self):
        """
        Whether object is immutable
        """

        return self._laia.is_frozen

    
    @property	
    def size(self):
        """
        Number of intervals in IntervalArray
        """

        return self._laia.size


    @property
    def range(self):
        """
        Ranges(start,  end) for each label
        """

        label_ranges = self._laia.label_ranges()
        if len(label_ranges) == 0:
            return None            
        
        return label_ranges["_IntervalArray"]

	
    @property
    def is_constructed(self):
        """
        Whether IntervalArray is constructed or not
        """

        return self._laia.is_constructed
	
    @property
    def starts(self):
        """
        Start values
        """

        starts = self._laia.starts

        return starts

	
    @property
    def ends(self):
        """
        End values
        """

        ends = self._laia.ends

        return ends

    def __len__(self):
        """
        Return size of IntervalArray
        """
        
        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        return self.size

	
    def __iter__(self):
        """
        Iterate over IntervalArray object
        """

        # Check if is constructed
        if self.is_constructed == False:
            self.construct()

        # Iterate through intervals
        for interval in self._laia:
            yield Interval(interval.start, interval.end)

	
    def __hash__(self):
        """
        Get hash value
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        return hash(self)


    def _wrap_laia(self, laia):
        """
        """

        self._laia.close()
        if isinstance(laia, LabeledInterval):
            laia = laia.to_array()
        self._laia = laia
    

    def __getitem__(self, key):
        """
        Index Intervals by value
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check that key given id 1D
        if isinstance(key, tuple) and len(key) == 2:
            raise IndexError("Incorrect number of dimensions given.")

        # Index
        if isinstance(key, int):
            new_labeled_intervals = self._laia.__getitem__(key)
            new_interval = Interval(new_labeled_intervals.start, new_labeled_intervals.end)

            return new_interval
        else:
            new_intervals = IntervalArray()
            new_labeled_intervals = self._laia.__getitem__(key)
            new_intervals._wrap_laia(new_labeled_intervals)

            return new_intervals


    def __repr__(self):
        """
        Representation of IntervalArray object
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Initialize string
        repr_string = "IntervalArray\n"

        # Iterate over labeled_interval_list
        if self.size > 10:
            for i in range(5):
                repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)
            repr_string += "   ...\n"
            for i in range(-5, 0, 1):
                repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)
        else:
            for i in range(self.size):
                repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)

        return repr_string


    def freeze(self):
        """
        Make :class:`~ailist.IntervalArray` immutable

        Parameters
        ----------
            None

        Returns
        -------
            None

        See Also
        --------
        IntervalArray.unfreeze: Make mutable
        IntervalArray.sort: Sort intervals inplace
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> ail = IntervalArray()
        >>> ail.add(1, 2)
        >>> ail.add(3, 4)
        >>> ail.add(3, 6)
        >>> ail
        IntervalArray
            (1-2)
            (3-4)
            (3-6)
        >>> ail.freeze()
        >>> ail.add(9, 10)
        TypeError: IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Make sure it is constructed
        if self.is_constructed == False:
            self.construct()

        # Change to frozen
        self.is_frozen = True


    def unfreeze(self):
        """
        Make :class:`~ailist.IntervalArray` mutable

        Parameters
        ----------
            None

        Returns
        -------
            None

        See Also
        --------
        IntervalArray.freeze: Make immutable
        IntervalArray.sort: Sort intervals inplace
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect

        Examples
        --------
        >>> from aiarray import IntervalArray
        >>> ail = IntervalArray()
        >>> ail.add(1, 2)
        >>> ail.add(3, 4)
        >>> ail.add(3, 6)
        >>> ail
        IntervalArray
            (1-2)
            (3-4)
            (3-6)
        >>> ail.freeze()
        >>> ail.add(9, 10)
        TypeError: IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.
        >>> ail.unfreeze()
        >>> ail.add(9, 10)

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Change to not frozen
        self.is_frozen = False


    def add(self, start, end):
        """
        Add an interval to IntervalArray inplace
        
        Parameters
        ----------
            start : int
                Start position of interval
            end : int
                End position of interval

        Returns
        -------
            None

        See Also
        --------
        IntervalArray.from_array: Add intervals from arrays
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
        IntervalArray.intersect: Find intervals overlapping given range

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> ail = IntervalArray()
        >>> ail.add(1, 2)
        >>> ail.add(3, 4)
        >>> ail.add(3, 6)
        >>> ail
        IntervalArray
            (1-2)
            (3-4)
            (3-6)

        """
        
        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check that object is not frozen
        if self.is_frozen:
            raise TypeError("IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.")

        # Add
        if isinstance(start, int):
            self._laia.add(start, end, "_IntervalArray")
        elif isinstance(start, np.ndarray):
            labels = np.repeat("_IntervalArray", len(start))
            self._laia.add(start, end, labels)

	
    def append(self, other_ail):
        """
        Add intervals from arrays to IntervalArray inplace
        
        Parameters
        ----------
            other_ail : IntervalArray
                Intervals to add to current IntervalArray

        Returns
        -------
            None

        See Also
        --------
        IntervalArray.add: Add interval to IntervalArray
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
        IntervalArray.intersect: Find intervals overlapping given range

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> import numpy as np
        >>> starts = np.arange(100)
        >>> ends = starts + 10
        >>> labels = np.repeat('a', len(starts))
        >>> ail = IntervalArray()
        >>> ail.from_array(starts, ends, labels)
        >>> ail
        IntervalArray
            (0-10, a)
            (1-11, a)
            (2-12, a)
            (3-13, a)
            (4-14, a)
            ...
            (95-105, a)
            (96-106, a)
            (97-107, a)
            (98-108, a)
            (99-109, a)
        
        >>> len(ail)
        100

        >>> ail2 = IntervalArray()
        >>> ail2.from_array(starts, ends, labels)
        >>> ail.append(ail2)
        >>> ail
        IntervalArray
            (0-10, a)
            (1-11, a)
            (2-12, a)
            (3-13, a)
            (4-14, a)
            ...
            (95-105, a)
            (96-106, a)
            (97-107, a)
            (98-108, a)
            (99-109, a)
        
        >>> len(ail)
        200

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check that object is not frozen
        if self.is_frozen:
            raise TypeError("IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.")
        
        self._laia.append(other_ail._laia)


    def construct(self, min_length = 20):
        """
        Construct labeled_aiarray_t *Required to call intersect

        Parameters
        ----------
            min_length : int
                Minimum length

        Returns
        -------
            None


        See Also
        --------
        IntervalArray.sort: Sort intervals inplace
        IntervalArray.intersect: Find intervals overlapping given range

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> ail = IntervalArray()
        >>> ail.add(1, 2)
        >>> ail.add(3, 4)
        >>> ail.add(2, 6)
        >>> ail
        IntervalArray
            (1-2)
            (3-4)
            (2-6)
        >>> ail.construct()
        >>> ail
        IntervalArray
            (1-2)
            (2-6)
            (3-4)

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("Labledaiarray object has been closed.")

        # Check if already constructed
        if self.is_constructed == False:
            self._laia.construct()
        else:
            pass


    def iter_sorted(self):
        """
        Iterate over an IntervalArray in sorted way
        
        Parameters
        ----------
            None
        
        Returns
        -------
            sorted_iter : Generator
                Generator of LabeledIntervals

        See Also
        --------
        IntervalArray.sort: Sort intervals inplace
        IntervalArray.intersect: Find intervals overlapping given range

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> ail = IntervalArray()
        >>> ail.add(1, 2)
        >>> ail.add(3, 4)
        >>> ail.add(2, 6)
        >>> ail
        LabledIntervalArray
            (1-2)
            (3-4)
            (2-6)
        >>> s_iter = ail.iter_sorted()
        >>> for i in s_iter:
        >>>		print(i)
        Interval(1-2)
        Interval(2-6)
        Interval(3-4)

        """

        # Check if is constructed
        if self.is_constructed == False:
            self.construct()

        # Iterate over labels in ail
        for i in self._laia.iter_sorted():
            interval = Interval(i.start, i.end)
            yield interval


    def iter_intersect(self,
                        query_laia,
                        return_intervals = True,
                        return_index = False):
        """
        """

        # Check if is constructed
        if self.is_constructed == False:
            self.construct()

        if return_intervals:
            if return_index:
                for overlaps, index in self._laia.iter_intersect(query_laia._laia,
                                                           return_intervals = True,
                                                           return_index = True):
                    i_overlaps = IntervalArray()
                    i_overlaps._wrap_laia(overlaps)
                    yield i_overlaps, index
            else:
                for overlaps in self.laia.iter_intersect(query_laia._laia,
                                                        return_intervals = True,
                                                        return_index = False):
                    i_overlaps = IntervalArray()
                    i_overlaps._wrap_laia(overlaps)
                    yield i_overlaps

        elif return_index:
            for index in self.iter_intersect(query_laia._laia,
                                             return_intervals = False,
                                             return_index = True):
                yield index


    def intersect(self, 
                    start, 
                    end, 
                    return_intervals = True,
                    return_index = False):
        """
        Find intervals overlapping given range
        
        Parameters
        ----------
            start : int | np.ndarray
                Start position of query range
            end : int | np.ndarray
                End position of query range

        Returns
        -------
            overlaps : IntervalArray
                Overlapping intervals

        .. warning::
            This requires :func:`~aiarray.IntervalArray.construct` and will run it if not already run. This will re-sort intervals inplace is ail.track_index = False.

        See Also
        --------
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
        IntervalArray.add: Add interval to IntervalArray
        IntervalArray.intersect_index: Find interval indices overlapping given range
        IntervalArray.intersect_from_array: Find interval indices overlapping given ranges

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> ail = IntervalArray()
        >>> ail.add(1, 2)
        >>> ail.add(3, 4)
        >>> ail.add(3, 6)
        >>> ail.add(3, 6)
        >>> ail
        IntervalArray
            (1-2)
            (3-4)
            (3-6)
            (3-6)
        >>> q = ail.intersect(2, 10)
        >>> q
        IntervalArray
            (3-4)
            (3-6)

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check if is constructed
        if self.is_constructed == False:
            self.construct()

        # Intersect
        if return_intervals:
            if return_index:
                label_overlaps, indices = self._laia.intersect(start, end, "_IntervalArray",
                                                               return_intervals=True,
                                                               return_index=True)
                overlaps = IntervalArray()
                overlaps._wrap_laia(label_overlaps)
                return overlaps, indices

            else:
                label_overlaps = self._laia.intersect(start, end, "_IntervalArray",
                                                      return_intervals=True,
                                                      return_index=False)
                overlaps = IntervalArray()
                overlaps._wrap_laia(label_overlaps)
                return overlaps

        elif return_index:
            indices = self._laia.intersect(start, end, "_IntervalArray",
                                            return_intervals=False,
                                            return_index=True)
            return indices

	
    def has_hit(self, start, end):
        """
        Find interval indices overlapping given ranges
        
        Parameters
        ----------
            starts : int | numpy.ndarray {long}
                Start positions of intervals
            ends : int | numpy.ndarray {long}
                End positions of intervals

        Returns
        -------
            has_hit : np.ndarray {bool}
                Bool array indicated overlap detected

        .. warning::
            This requires :func:`~aiarray.IntervalArray.construct` and will run it if not already run. This will re-sort intervals inplace if ail.track_index = False.

        See Also
        --------
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
        IntervalArray.add: Add interval to IntervalArray
        IntervalArray.intersect: Find intervals overlapping given range
        IntervalArray.intersect_index: Find interval indices overlapping given range

        Examples
        --------
        >>> from aiarray import IntervalArray
        >>> ail1 = IntervalArray()
        >>> ail1.add(1, 2)
        >>> ail1.add(3, 4)
        >>> ail1.add(2, 6)
        >>> ail1
        IntervalArray
            (1-2)
            (3-4)
            (2-6)
        >>> ail2 = IntervalArray()
        >>> ail2.add(1, 2)
        >>> ail2.add(3, 6)
        >>> ail2
        IntervalArray
            (1-2)
            (3-6)
        >>> q = ail1.intersect_from_array(ail2)
        >>> q
        (array([2, 1]), array([]))

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check if object is constructed
        if self.is_constructed == False:
            self.construct()

        # Initialize variables
        if isinstance(start, int):
            has_hit = self._laia.has_hit(start, end, "_IntervalArray")
        elif isinstance(start, np.ndarray):
            labels = np.repeat("_IntervalArray", len(start))
            has_hit = self._laia.has_hit(start, end, labels)
        
        return has_hit


    def intersect_from_IntervalArray(self,
                                     ail_query,
                                     return_intervals = True,
                                     return_index = False):
        """
        Find interval indices overlapping given ranges
        
        Parameters
        ----------
            ail_query : IntervalArray
                Intervals to query

        Returns
        -------
            ref_index : np.ndarray{int}
                Overlapping interval indices from IntervalArray
            query_index : np.ndarray{int}
                Overlapping interval indices from query IntervalArray

        See Also
        --------
        IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
        IntervalArray.add: Add interval to IntervalArray
        IntervalArray.intersect: Find intervals overlapping given range
        IntervalArray.intersect_from_array: Find interval indices overlapping given range

        Examples
        --------
        >>> from ailist import IntervalArray
        >>> ail1 = IntervalArray()
        >>> ail1.add(1, 2)
        >>> ail1.add(3, 4)
        >>> ail1.add(2, 6)
        >>> ail1
        IntervalArray
            range: (1-6)
            (1-2)
            (3-4)
            (2-6)
        >>> ail2 = IntervalArray()
        >>> ail2.add(1, 2)
        >>> ail2.add(3, 6)
        >>> ail2.add(3, 6)
        >>> ail2
        IntervalArray
            (1-2)
            (3-6)
        >>> q = ail1.intersect_from_IntervalArray(ail2)
        >>> q
        (array([0, 1, 1]), array([0, 2, 1]))

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check if object is constructed
        if self.is_constructed == False:
            self.construct()

        if return_intervals:
            if return_index:
                label_overlaps, indices = self._laia.intersect_from_LabeledIntervalArray(ail_query._laia,
                                                                                    return_intervals=True,
                                                                                    return_index=True)
                overlaps = IntervalArray()
                overlaps._wrap_laia(label_overlaps)
                return overlaps, indices
            else:
                label_overlaps = self._laia.intersect_from_LabeledIntervalArray(ail_query._laia,
                                                                                    return_intervals=True,
                                                                                    return_index=False)
                overlaps = IntervalArray()
                overlaps._wrap_laia(label_overlaps)
                return overlaps

        elif return_index:
            query_index, ref_index = self._laia.intersect_from_LabeledIntervalArray(ail_query._laia,
                                                                                    return_intervals=False,
                                                                                    return_index=True)
            return query_index, ref_index


    def nhits(self, start, end, min_length=None, max_length=None):
        """
        Find number of intervals overlapping given
        positions
        
        Parameters
        ----------
            starts : numpy.ndarray {long}
                Start positions to intersect
            ends : numpy.ndarray {long}
                End positions to intersect
            min_length : int
                Minimum length of intervals to include [default = None]
            max_length : int
                Maximum length of intervals to include [default = None]

        Returns
        -------
            nhits : numpy.ndarray{int}
                Number of hits per position

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("aiarray object has been closed.")

        # Make sure list is constructed
        if self.is_constructed == False:
            self.construct()

        if isinstance(start, int):
            nhits = self._laia.nhits(start, end, "_IntervalArray", max_length, min_length)
        elif isinstance(start, np.ndarray):
            labels = np.repeat("_IntervalArray", len(start))
            nhits = self._laia.nhits(start, end, labels, max_length, min_length)

        return nhits


    def bin_nhits(self, bin_size=100000, min_length=None, max_length=None):
        """
        Find number of intervals overlapping binned
        positions
        
        Parameters
        ----------
            bin_size : int
                Size of the bin to use
            min_length : int
                Minimum length of intervals to include [default = None]
            max_length : int
                Maximum length of intervals to include [default = None]

        Returns
        -------
            bins : pandas.Series {double}
                Position on index and coverage as values

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Determine bins
        label_bins = self._laia.bin_nhits(bin_size, min_length, max_length)
        bins = IntervalArray()
        bins._wrap_laia(label_bins)

        return bins

	
    def nhits_from_IntervalArray(self, query_laia, min_length=None, max_length=None):
        """
        Find number of intervals overlapping 
        
        Parameters
        ----------
            query_laia : IntervalArray
                Intervals to query
            min_length : int
                Minimum length of intervals to include [default = None]
            max_length : int
                Maximum length of intervals to include [default = None]

        Returns
        -------
            bins : pandas.Series {double}
                Position on index and coverage as values

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")
        if query_laia.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Initialize nhits
        nhits = self._laia.nhits_from_LabeledIntervalArray(query_laia, min_length, max_length)
        
        return nhits


    def wps(self, protection=60, min_length=None, max_length=None):
        """
        Calculate Window Protection Score
        for each position in IntervalArray range
        
        Parameters
        ----------
            protection : int
                Protection window to use
            min_length : int
                Minimum length of intervals to include [default = None]
            max_length : int
                Maximum length of intervals to include [default = None]

        Returns
        -------
            wps : pandas.Series {double}
                Position on index and WPS as values

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("aiarray object has been closed.")

        # Check if empty
        if self.laia.total_nr == 0:
                return None

        # Determine which labels to calculate
        wps_results = self._laia.wps(protection, "_IntervalArray", min_length, max_length)
        wps = wps_results["_IntervalArray"]
        
        return wps


    def coverage(self, min_length=None, max_length=None):
        """
        Calculate coverage
        for each position in IntervalArray range
        
        Parameters
        ----------
            min_length : int
                Minimum length of intervals to include [default = None]
            max_length : int
                Maximum length of intervals to include [default = None]

        Returns
        -------
            cov_results : dict {str:numpy.ndarray}
                Coverage as values

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("aiarray object has been closed.")

        # Check if empty
        if self.laia.total_nr == 0:
                return None

        # Determine which labels to calculate
        cov_results = self._laia.coverage("_IntervalArray", min_length, max_length)
        cov = cov_results["_IntervalArry"]
        
        return cov


    def merge(self, gap=0):
        """
        Merge intervals within a gap
        
        Parameters
        ----------
            gap : int
                Gap between intervals to merge

        Returns
        -------
            merged_list : IntervalArray
                Merged intervals

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("aiarray object has been closed.")

        # Make sure list is constructed
        if self.is_constructed == False:
            self.construct()

        # Create merged
        label_merged_list = self._laia.merge(gap)
        merged_list = IntervalArray()
        merged_list._wrap_laia(label_merged_list)

        return merged_list


    def filter(self, min_length=1, max_length=400):
        """
        Filter out intervals outside of a length range
        
        Parameters
        ----------
            min_length : int
                Minimum length to keep
            max_length : int
                Maximum langth to keep

        Returns
        -------
            filtered_ail : IntervalArray
                Filtered intervals

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Initialize filtered list
        label_filtered = self._laia.filter(min_length, max_length)
        filtered = IntervalArray()
        filtered._wrap_laia(label_filtered)

        return filtered


    def downsample(self,
                    proportion,
                    return_intervals = True,
                    return_index = True):
        """
        Randomly downsample IntervalArray
        
        Parameters
        ----------
            proportion : double
                Proportion of intervals to keep

        Returns
        -------
            filtered_ail : IntervalArray
                Downsampled IntervalArray

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        if return_intervals:
            if return_index:
                label_filtered, indices = self._laia.downsample(proportion,
                                                                return_intervals=True,
                                                                return_index=False)
                filtered = IntervalArray()
                filtered._wrap_laia(label_filtered)
                return filtered, indices
            else:
                label_filtered = self._laia.downsample(proportion,
                                                        return_intervals=True,
                                                        return_index=False)
                filtered = IntervalArray()
                filtered._wrap_laia(label_filtered)
                return filtered

        elif return_index:
            indices = self._laia.downsample(proportion, return_intervals=False, return_index=True)
            return indices


    def length_dist(self):
        """
        Calculate length distribution of intervals

        Parameters
        ----------
            None
        
        Returns
        -------
            distribution : numpy.ndarray{int}
                Interval length distribution

        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("AIList object has been closed.")


        # Calculate distribution
        distribution = self._laia.length_dist()

        return distribution


    def filter_exact_match(self, other_aiarray):
        """
        Determine which intervals are present
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("AIList object has been closed.")

        # Check construction
        if self.is_constructed == False:
            self.construct()
        if other_aiarray.is_constructed == False:
            other_aiarray.construct()

        label_matched_ail, indices = self._laia.filter_exact_match(other_aiarray._laia)
        matched_ail = IntervalArray()
        matched_ail._wrap_laia(label_matched_ail)

        return matched_ail, indices


    def has_exact_match(self, other_aiarray):
        """
        Determine which intervals are present
        """

        # Check if object is still open
        if self.is_closed or other_aiarray.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check if objects are constructed
        if self.is_constructed == False:
            self.construct()
        if other_aiarray.is_constructed == False:
            other_aiarray.construct()

        # Find matches
        has_match = self._laia.has_exact_match(other_aiarray._laia)

        return has_match


    def index_with_aiarray(self, other_aiarray):
        """
        """

        # Check if object is still open
        if self.is_closed or other_aiarray.is_closed:
            raise NameError("IntervalArray object has been closed.")

        return_code = self._laia.index_with_aiarray(other_aiarray._laia)

        if return_code == 1:
            raise NameError("Failed to run properly. Values are likely currupted now.")


    @staticmethod
    def create_bin(bin_range, bin_size=100000):
        """
        """

        label_range = {"_IntervalArray": bin_range}
        laia = LabeledIntervalArray.create_bin(label_range, bin_size)
        aia = IntervalArray()
        aia._wrap_laia(laia)

        return aia


    def simulate(self):
        """
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        label_simulation = self._laia.simulate()
        simulation = IntervalArray()
        simulation._wrap_laia(label_simulation)

        return simulation


    def sorted_index(self):
        """
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check if objects are constructed
        if self.is_constructed == False:
            self.construct()

        sorted_index = self._laia.sorted_index()

        return sorted_index


    def sort(self):
        """
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        # Check if objects are constructed
        if self.is_constructed == False:
            self.construct()

        self._laia.sort()

        return


    def validate_construction(self):
        """
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")
        
        res = self._laia.validate_construction()

        return res


    def copy(self):
        """
        Copy IntervalArray
        """

        # Check if object is still open
        if self.is_closed:
            raise NameError("IntervalArray object has been closed.")

        laia_copied = self._laia.copy()
        aia_copied = IntervalArray()
        aia_copied._wrap_laia(laia_copied)

        return aia_copied


    def close(self):
        """
        Close object and clear memory
        """

        # Free IntervalArray memory
        self._laia.close()